/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ukcprog.h>
#include <unistd.h>
#ifdef SOLARIS
#include <kstat.h>
#include <sys/sysinfo.h>
#endif
#ifdef FREEBSD
#include <sys/sysctl.h>
#include <sys/dkstat.h>
#endif
#define WAIT_TIME_IN_SECS 1
char *get_cpu_stats(){
		char *xml_cpu_stats;
		double u_p=0.00;
		double i_p=0.00;
		double k_p=0.00;
		double io_p=0.00;
		double s_p=0.00;
#ifdef LINUX
	long total, user, idle, kernel, nice;
	long cpu_states[4][2];
	FILE *f;
#endif
#ifdef SOLARIS
	kstat_ctl_t *kc;
	kstat_t *ksp;
	uint_t cpu_states[5][2];
	uint_t user, kernel, idle, iowait, swap, total;
	cpu_stat_t cs;
	
#endif
#ifdef FREEBSD
	long cp_time[CPUSTATES];
	long total, user, idle, kernel, nice;
	size_t size;	
#endif

#ifdef LINUX
	if ((f=fopen("/proc/stat", "r" ))==NULL) {	 
		errf("Failed to open /proc/stat (%m)");
		return NULL;
	}

	/* cpu stats should be the first line.. */
	if((fscanf(f,"%*s %ld %ld %ld %ld", &cpu_states[0][0], &cpu_states[1][0], &cpu_states[2][0] ,&cpu_states[3][0])) != 4){
		errf("Failed to read in correct number of CPU stats (%m)");
		return NULL;
	}

	if ((fclose(f)) != 0) {
		errf("Failed to close file (%m)");
		return NULL;
	}

	sleep(WAIT_TIME_IN_SECS);

	if ((f=fopen("/proc/stat", "r" ))==NULL) {	 
		errf("Failed to open /proc/stat (%m)");
		return NULL;
	}
	if((fscanf(f,"%*s %ld %ld %ld %ld", &cpu_states[0][1], &cpu_states[1][1], &cpu_states[2][1] ,&cpu_states[3][1])) != 4){
		errf("Failed to read in correct number of CPU stats (%m)");
		return NULL;
	}

	if ((fclose(f)) != 0) {
		errf("Failed to close file (%m)");
		return NULL;
	}

	user=cpu_states[0][1]-cpu_states[0][0]+cpu_states[1][1]-cpu_states[1][0];
	kernel=cpu_states[2][1]-cpu_states[2][0];
	idle=cpu_states[3][1]-cpu_states[3][0];
	total=user+kernel+idle;

	u_p=((double)user/(double)total)*100.00;
	k_p=((double)kernel/(double)total)*100.00;
	i_p=((double)idle/(double)total)*100.00;

#endif
#ifdef SOLARIS
        if ((kc = kstat_open()) == NULL) {
                errf("kstat_open failure (%m)");
                return NULL;
        }
	for (ksp = kc->kc_chain; ksp!=NULL; ksp = ksp->ks_next) {
		if ((strcmp(ksp->ks_module, "cpu_stat")) != 0) continue;
		if (kstat_read(kc, ksp, &cs) == -1) {
			errf("kstat read failure");
			continue;
		}
	}
	cpu_states[0][0]=cs.cpu_sysinfo.cpu[CPU_USER];
	cpu_states[1][0]=cs.cpu_sysinfo.cpu[CPU_WAIT];
	cpu_states[2][0]=cs.cpu_sysinfo.cpu[CPU_KERNEL];
	cpu_states[3][0]=cs.cpu_sysinfo.cpu[CPU_IDLE];
	cpu_states[4][0]=cs.cpu_sysinfo.cpu[CPU_STATES];

	sleep(WAIT_TIME_IN_SECS);
	
	if (kstat_chain_update(kc) == -1) {
		errf("Kstat update failure (%m)");
		return NULL;
	}
	for (ksp = kc->kc_chain; ksp!=NULL; ksp = ksp->ks_next) {
		if ((strcmp(ksp->ks_module, "cpu_stat")) != 0) continue;
		if (kstat_read(kc, ksp, &cs) == -1) {
			errf("kstat read failure");
			continue;
		}
	}
	cpu_states[0][1]=cs.cpu_sysinfo.cpu[CPU_USER];
	cpu_states[1][1]=cs.cpu_sysinfo.cpu[CPU_WAIT];
	cpu_states[2][1]=cs.cpu_sysinfo.cpu[CPU_KERNEL];
	cpu_states[3][1]=cs.cpu_sysinfo.cpu[CPU_IDLE];
	cpu_states[4][1]=cs.cpu_sysinfo.cpu[CPU_STATES];

	if((kstat_close(kc)) != 0){
		errf("Failed to close kstat control structure (%m)");
                return NULL;
        }

	user=cpu_states[0][1]-cpu_states[0][0];
	iowait=cpu_states[1][1]-cpu_states[1][0];
	kernel=cpu_states[2][1]-cpu_states[2][0];
	idle=cpu_states[3][1]-cpu_states[3][0];
	swap=cpu_states[4][1]-cpu_states[4][0];
	total=user+kernel+idle+iowait+swap;

	u_p=((double)user/(double)total)*100.00;
	k_p=((double)kernel/(double)total)*100.00;
	i_p=((double)idle/(double)total)*100.00;
	io_p=((double)iowait/(double)total)*100.00;
	s_p=((double)swap/(double)total)*100.00;

#endif	
#ifdef FREEBSD

#define CPU_STAT_NAME "kern.cp_time"

	if (sysctlbyname(CPU_STAT_NAME, NULL, &size, NULL, NULL) < 0){	
		errf("sysctrlbyname failed (%m)");
		return NULL;
	}
	if (sysctlbyname(CPU_STAT_NAME, &cp_time, &size, NULL, NULL) < 0){
		errf("Failed to get cpu stats (%m)");
		return NULL;
  	}

	user=cp_time[CP_USER];
  	nice=cp_time[CP_NICE];
  	kernel=cp_time[CP_SYS];
  	idle=cp_time[CP_IDLE];

	sleep(WAIT_TIME_IN_SECS);

	if (sysctlbyname(CPU_STAT_NAME, &cp_time, &size, NULL, NULL) < 0){
		errf("Failed to get cpu stats (%m)");
		return NULL;
  	}

	user=cp_time[CP_USER]-user;
  	nice=cp_time[CP_NICE]-nice;
  	kernel=cp_time[CP_SYS]-kernel;
  	idle=cp_time[CP_IDLE]-idle;
	total=user+nice+kernel+idle;

	u_p=((float)(user+nice)/(float)total)*100.0;
	k_p=((float)(kernel)/(float)total)*100.0;
	i_p=((float)(idle)/(float)total)*100.0;
	
#endif


	if((xml_cpu_stats=strf("<cpu><user>%3.2f</user><kernel>%3.2f</kernel><idle>%3.2f</idle><iowait>%3.2f</iowait><swap>%3.2f</swap></cpu>", u_p, k_p, i_p, io_p, s_p)) == NULL){
		errf("strf failed (%m)");
		return NULL;
	}

	return xml_cpu_stats;
}

