/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef LINUX
#include <string.h>
#endif
#ifdef SOLARIS
#include <unistd.h>
#include <kstat.h>
#endif
#ifdef FREEBSD
#include <sys/types.h>
#include <sys/sysctl.h>
#include <unistd.h>
#endif
#include "ukcprog.h"

char *get_memory_stats(){
	long long total=0;
	long long free=0;
	long long used=0;
	long long cache=0;
	char *xml_memory_stats;
#ifdef LINUX
	FILE *f;
	char *line;
	long long buffers;
#endif
#ifdef SOLARIS
	kstat_ctl_t *kc;
	kstat_t *ksp;
	kstat_named_t *kn;
	u_long totalmem;
	int pagesize;
#endif
#ifdef FREEBSD
#define TOTAL_MEM_NAME "hw.physmem"
#define FREE_MEM_NAME "vm.stats.vm.v_free_count"
#define INACTIVE_MEM_NAME "vm.stats.vm.v_inactive_count"
#define CACHE_MEM_NAME "vm.stats.vm.v_cache_count"
	int pagesize;
	long inactive;
	size_t size;
#endif

#ifdef LINUX
        if ((f=fopen("/proc/meminfo", "r" ))==NULL) {
                errf("Failed to open memory stats (%m)");
                return NULL;
        }
        while((line=fpgetline(f)) != NULL){
                if (((strncmp(line,"Mem:",4)) == 0)) {
                        if((sscanf(line, "%*s %lld %lld %lld %*s %lld %lld", &total, &used, &free, &buffers, &cache)) != 5){
                                errf("Failed to read memory details (%m)");
                                return NULL;
                        }
                        break;
                }
        }
        if ((fclose(f)) != 0) {
                errf("Failed to close file (%m)");
                return NULL;
        }

	cache=cache+buffers;
	used=total-free;

	total=(total/1024)/1024;
	used=(used/1024)/1024;
	free=(free/1024)/1024;
	cache=(cache/1024)/1024;
#endif
#ifdef SOLARIS
	if((pagesize=sysconf(_SC_PAGESIZE)) == -1){
		errf("Failed to get pagesize (%m)");
		return NULL;	
	}

	if((totalmem=sysconf(_SC_PHYS_PAGES)) == -1){
		errf("Failed to get total memory");
		return NULL;
	}

	if ((kc = kstat_open()) == NULL) {
		errf("kstat_open failure (%m)");
		return NULL;
	}
	if((ksp=kstat_lookup(kc, "unix", 0, "system_pages")) == NULL){
		errf("kstat lookup for \"system_pages\" (%m)");
		return NULL;
	}
	if (kstat_read(kc, ksp, 0) == -1) {
		errf("Failed to read kstat details (%m)");
		return NULL;
	}
	if((kn=kstat_data_lookup(ksp, "freemem")) == NULL){
		errf("Failed to get free memory (%m)");
		return NULL;
	}
        if((kstat_close(kc)) != 0){
                errf("Failed to close kstat control structure (%m)");
                return NULL;
        }	
	total=((totalmem*pagesize)/1024)/1024;
	free=(((kn->value.ul)*pagesize)/1024)/1024;
	used=total-free;
#endif
#ifdef FREEBSD
	if ((pagesize=getpagesize()) == -1){
		errf("Failed to get page size (%m)");
		return NULL;
	}

  	if (sysctlbyname(TOTAL_MEM_NAME, NULL, &size, NULL, NULL) < 0){
     		errf("sysctlbyname failed to get total memory (%m)");
		return NULL;
    	}
  	if (sysctlbyname(TOTAL_MEM_NAME, &total, &size, NULL, NULL) < 0){
    		errf("Failed to get total memory stats (%m)");
		return NULL;
  	}

  	if (sysctlbyname(FREE_MEM_NAME, NULL, &size, NULL, NULL) < 0){
     		errf("sysctlbyname failed to get free memory (%m)");
		return NULL;
    	}
  	if (sysctlbyname(FREE_MEM_NAME, &free, &size, NULL, NULL) < 0){
    		errf("Failed to get free memory stats (%m)");
		return NULL;
  	}

  	if (sysctlbyname(INACTIVE_MEM_NAME, NULL, &size, NULL, NULL) < 0){
     		errf("sysctlbyname failed to get inactive memory (%m)");
		return NULL;
    	}
  	if (sysctlbyname(INACTIVE_MEM_NAME, &inactive, &size, NULL, NULL) < 0){
    		errf("Failed to get inactive memory stats (%m)");
		return NULL;
  	}

	if (sysctlbyname(CACHE_MEM_NAME, NULL, &size, NULL, NULL) < 0){
     		errf("sysctlbyname failed to get cache memory (%m)");
		return NULL;
    	}
  	if (sysctlbyname(CACHE_MEM_NAME, &cache, &size, NULL, NULL) < 0){
    		errf("Failed to get cache memory stats (%m)");
		return NULL;
  	}

	total/=(1024*1024);
	free=((free*pagesize)/(1024*1024))+((inactive*pagesize)/(1024*1024));
	cache=((cache*pagesize)/(1024*1024));
	
	/* used should also be cache+active+wired but its less effort this way :) */
	used=total-free;
	  
#endif

	/* casting these from a long long to a long is a KLUDGE! please fix it :) */	
	if((xml_memory_stats=strf("<memory><total>%ld</total><free>%ld</free><used>%ld</used><cache>%ld</cache></memory>", (long)total, (long)free, (long)used, (long)cache)) == NULL){
		errf("strf failed (%m)");
		return NULL;
	}
	return xml_memory_stats;
}
