/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/utsname.h>
#include "ukcprog.h"
#include <stdio.h>
#ifdef SOLARIS
#include <kstat.h>
#include <time.h>
#endif
#ifdef FREEBSD
#include <sys/types.h>
#include <sys/sysctl.h>
#include <time.h>
#include <sys/time.h>
#endif

char *get_os_info(){
	char *xml_os_info;
	struct utsname os;
	long uptime;
#ifdef SOLARIS
	time_t boottime,curtime;
	kstat_ctl_t *kc;
	kstat_t *ksp;
	kstat_named_t *kn;
#endif
#ifdef LINUX
	FILE *f;
#endif
#ifdef FREEBSD
	struct timeval boottime;
	time_t curtime;
	size_t size;
#endif
	if((uname(&os)) < 0){
		errf("Failed to get os stats (%m)");
		return NULL;
	}

	/* Get the uptime */
#ifdef SOLARIS
	if ((kc = kstat_open()) == NULL) {
		errf("kstat_open failure (%m)");
		return NULL;
	}
	if((ksp=kstat_lookup(kc, "unix", -1, "system_misc"))==NULL){
		errf("failed to find lookup information (%m)");
		return NULL;
	}
	if (kstat_read(kc, ksp, 0) == -1) {
		errf("Failed to read kernel information (%m)");
		return NULL;
	}
	if((kn=kstat_data_lookup(ksp, "boot_time")) == NULL){
		errf("Failed to get uptime (%m)");
		return NULL;
	}
	boottime=(kn->value.ui32);
	time(&curtime);
	if((kstat_close(kc)) != 0){
		errf("Failed to close kstat control structure (%m)");
		return NULL;
	}
	uptime=curtime-boottime;
#endif
#ifdef LINUX
	if ((f=fopen("/proc/uptime", "r")) == NULL) {
		errf("Failed to open /proc/uptime (%m)");
		return NULL;
	}
	if((fscanf(f,"%ld",&uptime)) != 1){
		errf("Failed to read uptime (%m)");
		return NULL;
	}
	if ((fclose(f)) != 0) {
		errf("Failed to close file (%m)");
		return NULL;
	}
#endif
#ifdef FREEBSD
	#define BOOT_TIME_NAME "kern.boottime"

	if (sysctlbyname(BOOT_TIME_NAME, NULL, &size, NULL, NULL) < 0){
		errf("sysctlbyname failed (%m)");
		return NULL;
	}
	if (sysctlbyname(BOOT_TIME_NAME, &boottime, &size, NULL, NULL) < 0){
		errf("sysctlbyname failed (%m)");
		return NULL;
	}
	time(&curtime);
	uptime=curtime-boottime.tv_sec;
	
#endif


	xml_os_info=strf("<os><name>%s</name><release>%s</release><version>%s</version><sysname>%s</sysname><platform>%s</platform><uptime>%ld</uptime></os>", os.sysname, os.release, os.version, os.nodename, os.machine, uptime);

	return xml_os_info;
}
