/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include "ukcprog.h"
#include <unistd.h>
#ifdef SOLARIS
#include <kstat.h>
#include <sys/sysinfo.h>
#endif
#ifdef LINUX
#include <string.h>
#endif
#ifdef FREEBSD
#include <sys/types.h>
#include <sys/sysctl.h>
#endif

#define WAIT_TIME_IN_SECS 1

char *get_page_stats(){
	char *xml_page_stats;
	int swap_pageins;
	int swap_pageouts;
#ifdef SOLARIS
        kstat_ctl_t *kc;
        kstat_t *ksp;
        cpu_stat_t cs;
	uint_t swapin, swapout;
#endif
#ifdef LINUX
        FILE *f;
        char *line;
	long  swapin[2], swapout[2];
#endif
#ifdef FREEBSD
	size_t size;
	long swapin[2], swapout[2];
#endif


#ifdef SOLARIS
        if ((kc = kstat_open()) == NULL) {
                errf("kstat_open failure (%m)");
                return NULL;
        }
        for (ksp = kc->kc_chain; ksp!=NULL; ksp = ksp->ks_next) {
                if ((strcmp(ksp->ks_module, "cpu_stat")) != 0) continue;
                if (kstat_read(kc, ksp, &cs) == -1) {
                        errf("kstat read failure");
                        continue;
                }
        }

	swapin=cs.cpu_vminfo.pgswapin;
	swapout=cs.cpu_vminfo.pgswapout;

	sleep(WAIT_TIME_IN_SECS);

        if (kstat_chain_update(kc) == -1) {
                errf("Kstat update failure (%m)");
                return NULL;
        }
        for (ksp = kc->kc_chain; ksp!=NULL; ksp = ksp->ks_next) {
                if ((strcmp(ksp->ks_module, "cpu_stat")) != 0) continue;
                if (kstat_read(kc, ksp, &cs) == -1) {
                        errf("kstat read failure");
                        continue;
                }
        }
        if((kstat_close(kc)) != 0){
                errf("Failed to close kstat control structure (%m)");
                return NULL;
        }
	swap_pageins=(cs.cpu_vminfo.pgswapin-swapin)/WAIT_TIME_IN_SECS;
	swap_pageouts=(cs.cpu_vminfo.pgswapout-swapout)/WAIT_TIME_IN_SECS;
#endif
#ifdef LINUX
       if ((f=fopen("/proc/stat", "r" ))==NULL) {
                errf("Failed to open /proc/stat (%m)");
                return NULL;
        }
        while((line=fpgetline(f)) != NULL){
                if (((strncmp(line,"swap",4)) == 0)) {
                        if((sscanf(line, "%*s %ld %ld", &swapin[0], &swapout[0])) != 2){
                                errf("Failed to read swap paging details (%m)");
                                return NULL;
                        }
                        break;
                }
        }
        if ((fclose(f)) != 0) {
                errf("Failed to close file (%m)");
                return NULL;
        }
	sleep(WAIT_TIME_IN_SECS);

        if ((f=fopen("/proc/stat", "r" ))==NULL) {
                errf("Failed to open /proc/stat (%m)");
                return NULL;
        }
        while((line=fpgetline(f)) != NULL){
                if (((strncmp(line,"swap",4)) == 0)) {
                        if((sscanf(line, "%*s %ld %ld", &swapin[1], &swapout[1])) != 2){
                                errf("Failed to read swap paging details (%m)");
                                return NULL;
                        }
                        break;
                }
        }
        if ((fclose(f)) != 0) {
                errf("Failed to close file (%m)");
                return NULL;
        }

	swap_pageins=(swapin[1]-swapin[0])/WAIT_TIME_IN_SECS;
	swap_pageouts=(swapout[1]-swapout[0])/WAIT_TIME_IN_SECS;
#endif
#ifdef FREEBSD
#define SWAP_IN_NAME "vm.stats.vm.v_swappgsin"
#define SWAP_OUT_NAME "vm.stats.vm.v_swappgsout"
        if (sysctlbyname(SWAP_IN_NAME, NULL, &size, NULL, NULL) < 0){
                errf("sysctlbyname failed to get total memory (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_IN_NAME, &swapin[0], &size, NULL, NULL) < 0){
                errf("Failed to get total memory stats (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_OUT_NAME, NULL, &size, NULL, NULL) < 0){
                errf("sysctlbyname failed to get total memory (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_OUT_NAME, &swapout[0], &size, NULL, NULL) < 0){
                errf("Failed to get total memory stats (%m)");
                return NULL;
        }

	sleep(WAIT_TIME_IN_SECS);

        if (sysctlbyname(SWAP_IN_NAME, NULL, &size, NULL, NULL) < 0){
                errf("sysctlbyname failed to get total memory (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_IN_NAME, &swapin[1], &size, NULL, NULL) < 0){
                errf("Failed to get total memory stats (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_OUT_NAME, NULL, &size, NULL, NULL) < 0){
                errf("sysctlbyname failed to get total memory (%m)");
                return NULL;
        }
        if (sysctlbyname(SWAP_OUT_NAME, &swapout[1], &size, NULL, NULL) < 0){
                errf("Failed to get total memory stats (%m)");
                return NULL;
        }

	swap_pageins=(swapin[1]-swapin[0])/WAIT_TIME_IN_SECS;
	swap_pageouts=(swapout[1]-swapout[0])/WAIT_TIME_IN_SECS;

#endif
	

	if((xml_page_stats=strf("<pages><swapins>%d</swapins><swapouts>%d</swapouts></pages>", swap_pageins, swap_pageouts)) == NULL){
		errf("strf failed (%m)");
		return NULL;
	}
	return xml_page_stats;
}
